use super::Error;
use pest::{Parser, Span};
use pest_derive::Parser;
use std::str::FromStr;

#[derive(Parser)]
#[grammar = "src/filter_expr/expr.pest"] // relative path to your .pest file
struct QueryParser;

#[derive(Clone, PartialEq, Eq, Debug, Copy)]
pub enum ExprOp {
    And,
    Or,
    Not,
    Eq,
    NotEq,
    Gt,
    Gte,
    Lt,
    Lte,
    Add,
    Sub,
    Mul,
    Div,
}

#[derive(Clone, PartialEq, Eq, Debug, Copy)]
pub enum Order {
    Ascending,
    Descending,
}

impl FromStr for Order {
    type Err = Error;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        match s {
            "asc" => Ok(Order::Ascending),
            "desc" => Ok(Order::Descending),
            s => Err(Error::InvalidOrder(s.to_owned())),
        }
    }
}

impl From<&str> for ExprOp {
    fn from(s: &str) -> Self {
        match s.to_lowercase().as_str() {
            "+" => ExprOp::Add,
            "-" => ExprOp::Sub,
            "*" => ExprOp::Mul,
            "/" => ExprOp::Div,
            "&&" | "and" => ExprOp::And,
            "||" | "or" => ExprOp::Or,
            "!" | "not" => ExprOp::Not,
            "==" | "=" => ExprOp::Eq,
            "!=" => ExprOp::NotEq,
            ">" => ExprOp::Gt,
            ">=" => ExprOp::Gte,
            "<" => ExprOp::Lt,
            "<=" => ExprOp::Lte,
            x => panic!("{} is not a valid Operation", x),
        }
    }
}

#[derive(Debug)]
enum Expr {
    Not(Box<Expr>),
    Op {
        op: ExprOp,
        left: Box<Expr>,
        right: Box<Expr>,
    },
    Variable(Vec<String>),
    String(String),
    Ident(String),
    Number(i128),
}

#[derive(Debug)]
pub struct Query {
    where_clause: Option<Expr>,
    limit_clause: Option<i128>,
    order_by_clause: Option<(Expr, Order)>,
}

/// Borrowed from https://github.com/pest-parser/pest/blob/master/meta/src/parser.rs#L687C2-L741C2
fn unescape(string: &str) -> Option<String> {
    let mut result = String::new();
    let mut chars = string.chars();

    loop {
        match chars.next() {
            Some('\\') => match chars.next()? {
                '"' => result.push('"'),
                '\\' => result.push('\\'),
                'r' => result.push('\r'),
                'n' => result.push('\n'),
                't' => result.push('\t'),
                '0' => result.push('\0'),
                '\'' => result.push('\''),
                'x' => {
                    let string: String = chars.clone().take(2).collect();
                    if string.len() != 2 {
                        return None;
                    }
                    for _ in 0..string.len() {
                        chars.next()?;
                    }
                    let value = u8::from_str_radix(&string, 16).ok()?;
                    result.push(char::from(value));
                }
                'u' => {
                    if chars.next()? != '{' {
                        return None;
                    }
                    let string: String = chars.clone().take_while(|c| *c != '}').collect();
                    if string.len() < 2 || 6 < string.len() {
                        return None;
                    }
                    for _ in 0..string.len() + 1 {
                        chars.next()?;
                    }
                    let value = u32::from_str_radix(&string, 16).ok()?;
                    result.push(char::from_u32(value)?);
                }
                _ => return None,
            },
            Some(c) => result.push(c),
            None => return Some(result),
        };
    }
}

fn parse_expr(pair: pest::iterators::Pair<Rule>) -> Result<Expr, Error> {
    Ok(match pair.as_rule() {
        Rule::logical_expr
        | Rule::logical_term
        | Rule::comparison
        | Rule::sum
        | Rule::factor
        | Rule::product => {
            let mut inner = pair.into_inner();
            let mut last_expr = None;

            while let Some(first) = inner.next() {
                let (first_expr, next) = if let Some(first_expr) = last_expr {
                    (first_expr, Some(first))
                } else {
                    (parse_expr(first)?, inner.next())
                };
                last_expr = Some(if let Some(op_pair) = next {
                    let op = op_pair.as_str().into();
                    let second = inner.next().ok_or(Error::MissingNextRule)?;
                    let second_expr = parse_expr(second)?;
                    Expr::Op {
                        op,
                        left: Box::new(first_expr),
                        right: Box::new(second_expr),
                    }
                } else {
                    first_expr
                });
            }

            last_expr.ok_or(Error::MissingNextRule)?
        }
        Rule::string => {
            let mut inner = pair.into_inner();
            let _ = inner.next();

            Expr::String(
                inner
                    .next()
                    .map(|x| unescape(x.as_str()))
                    .flatten()
                    .expect("invalid escape"),
            )
        }
        Rule::not_expr => {
            let mut inner = pair.into_inner();
            Expr::Not(Box::new(
                inner
                    .next()
                    .map(parse_expr)
                    .ok_or(Error::MissingNextRule)??,
            ))
        }
        Rule::variable => {
            let mut inner = pair.into_inner();
            let mut first = vec![inner
                .next()
                .ok_or(Error::MissingNextRule)?
                .as_str()
                .to_owned()];
            while let Some(Ok(Expr::Ident(ident))) = inner.next().map(parse_expr) {
                first.push(ident);
            }
            Expr::Variable(first)
        }
        Rule::number => Expr::Number(pair.as_str().parse()?),
        Rule::ident => Expr::Ident(pair.as_str().to_owned()),
        rule => return Err(Error::UnexpectedRule(rule)),
    })
}

pub fn parse_query(query: &str) -> Result<Query, Error> {
    let pairs = QueryParser::parse(Rule::query, query)?;

    let mut where_clause = None;
    let mut limit_clause = None;
    let mut order_by_clause = None;

    for pair in pairs {
        match pair.as_rule() {
            Rule::EOI => {
                break;
            }
            Rule::where_clause => {
                where_clause = Some(
                    pair.into_inner()
                        .next()
                        .map(parse_expr)
                        .ok_or(Error::MissingNextRule)??,
                );
            }
            Rule::limit_clause => {
                limit_clause = match pair
                    .into_inner()
                    .next()
                    .map(parse_expr)
                    .ok_or(Error::MissingNextRule)??
                {
                    Expr::Number(v) => Some(v),
                    _ => unreachable!(),
                };
            }
            Rule::order_by_clause => {
                let mut iter = pair.into_inner();
                let expr = iter
                    .next()
                    .map(parse_expr)
                    .ok_or(Error::MissingNextRule)??;
                let order = if let Some(order_by_text) = iter.next() {
                    order_by_text.as_str().to_lowercase().parse()?
                } else {
                    Order::Ascending
                };

                order_by_clause = Some((expr, order));
            }
            _ => unreachable!(),
        }
    }

    Ok(Query {
        where_clause,
        limit_clause,
        order_by_clause,
    })
}
